/* 
 * Copyright (C) 2002 Laird Breyer
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * Author:   Laird Breyer <laird@lbreyer.com>
 */

/* we define several memory models, which differ basically 
   in the number of bytes used for the hash tables. Adjust to taste */

/* use this for 64-bit hashes */
#undef HUGE_MEMORY_MODEL
/* use this for 32-bit hashes */
#define NORMAL_MEMORY_MODEL 
/* use this for 16-bit hashes */
#undef SMALL_MEMORY_MODEL
/* use this for 8-bit hashes */
#undef TINY_MEMORY_MODEL

/* the following defines set up a tradeoff between
   modelling accuracy and memory requirements - season to taste */

/* digram digitization: avg loss of precision = 0.01 * token size */
#define DIGITIZE_DIGRAMS
/* lambda digitization: avg loss of precision = 0.01 */
#define DIGITIZE_LAMBDA
/* learner.hash digitization: avg loss of precision = 0.01 */
#define DIGITIZE_LWEIGHTS
/* tolerance for the error in divergence - 
   more accuracy means slower learning. */
#define TOL 0.05

#if defined HAVE_MBRTOWC

#include <wctype.h>
#include <wchar.h>

#else

#undef HAVE_LIBBOOST_REGEX

#endif

#include <limits.h>
#include <stdio.h>

#if defined HAVE_LIBBOOST_REGEX

#define UNICODE
#include <boost/regex.h>

#else 

#include <sys/types.h>
#include <regex.h>

#endif

/* some systems seem to have broken sys/types */
#if defined OS_SUN

#include <ieeefp.h>

typedef uint8_t u_int8_t;
typedef uint16_t u_int16_t;
typedef uint32_t u_int32_t;
typedef uint64_t u_int64_t;

#endif


/* below, FMT_* macros are used in printf/scanf format strings */
#if defined HUGE_MEMORY_MODEL

typedef u_int64_t token_count_t;
typedef u_int8_t token_order_t;
typedef u_int8_t hash_bit_count_t;
typedef u_int64_t hash_value_t;
typedef u_int64_t hash_count_t;
typedef unsigned int hash_percentage_t;
typedef u_int8_t category_count_t;
typedef u_int8_t regex_count_t;
typedef u_int32_t document_count_t;

typedef int16_t digitized_weight_t;
typedef float weight_t;
typedef long double score_t;
#define FMT_printf_score_t "Lf"
#define FMT_scanf_score_t "Lf"

typedef u_int8_t token_stack_t;
typedef int charbuf_len_t;
typedef u_int16_t alphabet_size_t;
typedef u_int16_t smbitmap_t;

typedef int error_code_t;
typedef int bool_t;

/* where token counts wrap around */
#define K_TOKEN_COUNT_MAX ((token_count_t)18446744073709551615U)
/* size of hash in bits */
#define MAX_HASH_BITS ((hash_bit_count_t)64)
/* maximum size of a token, beyond that rest is ignored */
#define MAX_TOKEN_LEN ((charbuf_len_t)100)
/* for line filtering: maximum number of tokens allowed on a single line */
#define MAX_TOKEN_LINE_STACK ((token_stack_t)256) 
#define BUFLEN ((charbuf_len_t)1024)

#elif defined NORMAL_MEMORY_MODEL

typedef u_int32_t token_count_t;
typedef u_int8_t token_order_t;
typedef u_int8_t hash_bit_count_t;
typedef u_int32_t hash_value_t;
typedef u_int32_t hash_count_t;
typedef unsigned int hash_percentage_t;
typedef u_int8_t category_count_t;
typedef u_int8_t regex_count_t;
typedef u_int32_t document_count_t;

typedef int16_t digitized_weight_t;
typedef float weight_t;
typedef double score_t;
#define FMT_printf_score_t "f"
#define FMT_scanf_score_t "lf"

typedef u_int8_t token_stack_t;
typedef int charbuf_len_t;
typedef u_int16_t alphabet_size_t;
typedef u_int16_t smbitmap_t;

typedef int error_code_t;
typedef int bool_t;



/* where token counts wrap around */
#define K_TOKEN_COUNT_MAX ((token_count_t)4294967295U)
/* size of hash in bits */
#define MAX_HASH_BITS ((hash_bit_count_t)30)
/* maximum size of a token, beyond that rest is ignored */
#define MAX_TOKEN_LEN ((charbuf_len_t)100)
/* for line filtering: maximum number of tokens allowed on a single line */
#define MAX_TOKEN_LINE_STACK ((token_stack_t)128) 
#define BUFLEN ((charbuf_len_t)1024)

#elif defined SMALL_MEMORY_MODEL

typedef u_int32_t token_count_t;
typedef u_int8_t token_order_t;
typedef u_int8_t hash_bit_count_t;
typedef u_int16_t hash_value_t;
typedef u_int16_t hash_count_t;
typedef unsigned int hash_percentage_t;
typedef u_int8_t category_count_t;
typedef u_int8_t regex_count_t;
typedef u_int16_t document_count_t;

typedef int16_t digitized_weight_t;
typedef float weight_t;
typedef double score_t;
#define FMT_printf_score_t "f"
#define FMT_scanf_score_t "lf"

typedef u_int8_t token_stack_t;
typedef int charbuf_len_t;
typedef u_int16_t alphabet_size_t;
typedef u_int16_t smbitmap_t;

typedef int error_code_t;
typedef int bool_t;

/* where token counts wrap around */
#define K_TOKEN_COUNT_MAX ((token_count_t)4294967295U)
/* size of hash in bits */
#define MAX_HASH_BITS ((hash_bit_count_t)15)
/* maximum size of a token, beyond that rest is ignored */
#define MAX_TOKEN_LEN ((charbuf_len_t)100)
/* for line filtering: maximum number of tokens allowed on a single line */
#define MAX_TOKEN_LINE_STACK ((token_stack_t)128) 
#define BUFLEN ((charbuf_len_t)1024)

#elif defined TINY_MEMORY_MODEL
/* not tested, this model probably doesn't work ;-) */
#undef DIGITIZE_DIGRAMS

typedef u_int32_t token_count_t;
typedef u_int8_t token_order_t;
typedef u_int8_t hash_bit_count_t;
typedef u_int8_t hash_value_t;
typedef u_int8_t hash_count_t;
typedef unsigned int hash_percentage_t;
typedef u_int8_t category_count_t;
typedef u_int8_t regex_count_t;
typedef u_int8_t document_count_t;

typedef int16_t digitized_weight_t;
typedef float weight_t;
typedef double score_t;
#define FMT_printf_score_t "f"
#define FMT_scanf_score_t "lf"

typedef u_int8_t token_stack_t;
typedef int charbuf_len_t;
typedef u_int16_t alphabet_size_t;
typedef u_int16_t smbitmap_t;

typedef int error_code_t;
typedef int bool_t;

#define K_TOKEN_COUNT_MAX ((token_count_t)4294967295U)
/* size of hash in bits */
#define MAX_HASH_BITS ((hash_bit_count_t)8)
/* maximum size of a token, beyond that rest is ignored */
#define MAX_TOKEN_LEN ((charbuf_len_t)100)
/* for line filtering: maximum number of tokens allowed on a single line */
#define MAX_TOKEN_LINE_STACK ((token_stack_t)128) 
#define BUFLEN ((charbuf_len_t)512)

#endif

/* this is common to all memory models */

#if defined OS_DARWIN
/* the system I tested this on didn't seem to like packed structures */
#define PACK_STRUCTS

#else

/* disable this if speed is paramount  */
#define PACK_STRUCTS __attribute__ ((packed))

#endif

/* when digitizing transitions, this stands for -infinity */
#define DIGITIZED_WEIGHT_MIN ((digitized_weight_t)SHRT_MIN)
#define DIGITIZED_WEIGHT_MAX ((digitized_weight_t)SHRT_MAX)
/* maximum number of categories we can handle simultaneously */
#define MAX_CAT ((category_count_t)64)
/* percentage of hash we use */
#define HASH_FULL ((hash_percentage_t)95)
/* alphabet size */
#define ASIZE ((alphabet_size_t)256)
/* we need two special markers, which cannot be part 
 * of the alphabet. Fortunately, we can use ASCII control
 * characters. Hopefully, these won't be used for anything important. 
 */
#define DIAMOND '\001'
#define TOKENSEP '\002'
#define AMIN '\003'

/* user options */
#define U_OPTION_CLASSIFY               1
#define U_OPTION_LEARN                  2
#define U_OPTION_FASTEMP                3
#define U_OPTION_CUTOFF                 4
#define U_OPTION_VERBOSE                5
#define U_OPTION_STDIN                  6
#define U_OPTION_SCORES                 7
#define U_OPTION_POSTERIOR              8
#define U_OPTION_FILTER                 9
#define U_OPTION_DEBUG                  10
#define U_OPTION_DUMP                   12
#define U_OPTION_APPEND                 13
#define U_OPTION_DECIMATE               14
#define U_OPTION_GROWHASH               15
#define U_OPTION_INDENTED               16

/* model options */
#define M_OPTION_REFMODEL               1
#define M_OPTION_TEXT_FORMAT            2
#define M_OPTION_MBOX_FORMAT            3
#define M_OPTION_XML                    4
#define M_OPTION_I18N                   5
#define M_OPTION_CASEN                  6
#define M_OPTION_CALCENTROPY            7
#define M_OPTION_MULTINOMIAL            8
#define M_OPTION_LAPLACE                9
#define M_OPTION_CHAR_ALPHA             10
#define M_OPTION_CHAR_ALNUM             11
#define M_OPTION_CHAR_GRAPH             12
#define M_OPTION_HEADERS                13
#define M_OPTION_PLAIN                  14
#define M_OPTION_NOPLAIN                15
#define M_OPTION_SHOW_LINKS             16
#define M_OPTION_SHOW_ALT               17
#define M_OPTION_HTML                   18
#define M_OPTION_XHEADERS               19
#define M_OPTION_CHAR_CEF               20
#define M_OPTION_SHOW_SCRIPT            21
#define M_OPTION_SHOW_HTML_COMMENTS     22
#define M_OPTION_NOREGEX                23
#define M_OPTION_ATTACHMENTS            24

typedef u_int32_t options_t; /* make sure big enough for all options */
#define FMT_printf_options_t "d"
#define FMT_scanf_options_t "ld"

typedef long int re_bitfield;
/* maximum number of regular expressions we can handle */
#define MAX_RE ((regex_count_t)(8 * sizeof(re_bitfield)))
/* maximum number of tagged subexpressions we can handle for each regex */
#define MAX_SUBMATCH ((token_order_t)10)

/* macros */

/* used for digitizing */
#if defined DIGITIZE_LWEIGHTS

#define PACK_LWEIGHTS(a) ((digitized_weight_t)(digitize_a_weight(a)))
#define UNPACK_LWEIGHTS(a) ((weight_t)(a)/100.00)

#else

#define PACK_LWEIGHTS(a) ((weight_t)(a))
#define UNPACK_LWEIGHTS(a) ((weight_t)(a))

#endif

#if defined DIGITIZE_LAMBDA

#define PACK_LAMBDA(a) ((digitized_weight_t)(digitize_a_weight(a)))
#define UNPACK_LAMBDA(a) ((weight_t)(a)/100.00)

#else

#define PACK_LAMBDA(a) ((weight_t)(a))
#define UNPACK_LAMBDA(a) ((weight_t)(a))

#endif

#if defined DIGITIZE_DIGRAMS

#define PACK_DIGRAMS(a) ((digitized_weight_t)(digitize_a_weight(a)))
#define UNPACK_DIGRAMS(a) ((weight_t)(a)/100.00)
#define SIZEOF_DIGRAMS (sizeof(digitized_weight_t))

#else

#define PACK_DIGRAMS(a) ((weight_t)(a))
#define UNPACK_DIGRAMS(a) ((weight_t)(a))
#define SIZEOF_DIGRAMS (sizeof(weight_t))

#endif

/* used in hash code */

#define FILLEDP(a) ((a)->id)
#define EQUALP(a,b) ((a)==(b))
#define SET(a,b) (a = (b))

#define SETMARK(a) ((a)->ltrms = 1.0)
#define UNSETMARK(a) ((a)->ltrms = 0.0)
#define MARKEDP(a) ((a)->ltrms == 1.0)

/* used by both category load and learner save functions */
#define MAGIC_BUFSIZE 256
#define MAGIC1    "# dbacl version "VERSION" category %s %s\n"
#define MAGIC2_i  "# entropy %" FMT_scanf_score_t \
                  " logZ %" FMT_scanf_score_t " max_order %hd" \
                  " type %s\n"
#define MAGIC2_o  "# entropy %" FMT_printf_score_t \
                  " logZ %" FMT_printf_score_t " max_order %hd" \
                  " type %s\n"
#define MAGIC3    "# hash_size %hd" \
                  " features %ld unique_features %ld" \
                  " documents %ld\n"
#define MAGIC4_i  "# options %" FMT_scanf_options_t "\n"
#define MAGIC4_o  "# options %" FMT_printf_options_t "\n"
#define MAGIC5_i  "# regex %s\n"
#define MAGIC5_o  "# regex %s||%s\n"
#define MAGIC5_wo "# regex %ls||%s\n"
#define RESTARTPOS 8
#define MAGIC6    "#\n"

/* data structures */

typedef struct {
  hash_value_t id;
  token_count_t count;
} h_item;

typedef struct {
  hash_count_t max_tokens;
  hash_bit_count_t max_hash_bits;
  token_count_t full_token_count;
  token_count_t unique_token_count;
  h_item *hash;
  bool_t track_features;
  h_item *feature_stack[MAX_TOKEN_LINE_STACK];
  token_stack_t feature_stack_top;
} Empirical;

typedef struct {
  hash_value_t id;
#if defined DIGITIZE_LAMBDA
  digitized_weight_t lam;
#else
  weight_t lam;
#endif
} PACK_STRUCTS c_item;

typedef struct {
  char *filename;
  enum {simple, sequential} model_type;
  token_order_t max_order;
  token_count_t complexity;
  token_count_t model_unique_token_count;
  token_count_t model_full_token_count;
  document_count_t model_num_docs;
  hash_count_t max_tokens;
  hash_bit_count_t max_hash_bits;
  re_bitfield retype;
  score_t logZ;
  score_t divergence;
  score_t score;
  c_item *hash;
#if defined DIGITIZE_DIGRAMS
  digitized_weight_t dig[ASIZE][ASIZE];
#else
  weight_t dig[ASIZE][ASIZE];
#endif
} Category;

typedef struct {
  token_count_t count;
  weight_t lam; 
#if defined DIGITIZE_LWEIGHTS
  digitized_weight_t ltrms;
  digitized_weight_t dref;
#else
  weight_t ltrms;
  weight_t dref;
#endif
  hash_value_t id;
  token_order_t order;
} PACK_STRUCTS l_item;

typedef struct {
  char *filename;
  FILE *tmp;
  token_order_t max_order;
  token_count_t fixed_order_token_count[MAX_SUBMATCH];
  token_count_t fixed_order_unique_token_count[MAX_SUBMATCH];
  hash_bit_count_t max_hash_bits;
  hash_count_t max_tokens;
  token_count_t full_token_count;
  token_count_t unique_token_count;
  score_t logZ;
  score_t divergence;
  l_item *hash;
  weight_t dig[ASIZE][ASIZE];
  long int regex_token_count[MAX_RE + 1];
  document_count_t num_docs;
} Learner;

typedef struct {
  double alpha;
  double u[ASIZE];
} Dirichlet;

typedef struct {
  regex_t regex;
#if defined HAVE_LIBBOOST_REGEX
  wchar_t *string;
#else 
  char *string;
#endif
  smbitmap_t submatches;
} Regex;

#define MAX_BOUNDARIES 8

#define MAX_BOUNDARY_BUFSIZE 70

typedef struct {
  enum { ctUNDEF, ctPLAINTEXT, ctRICHTEXT, ctHTML, ctXML, ctSGML, ctUNTEXT, ctOTHER, ctAPP } type;
  enum { ceUNDEF, ceID, ceB64, ceQP, ceBIN } encoding;
} MIME_Struct;

typedef enum { htSTANDARD, htEXTENDED, htMIME, htCONT, htUNDEF } HEADER_Type;

typedef struct {
  enum { msUNDEF, msHEADER, msBODY, msATTACH} state;
  enum { msuUNDEF, msuTRACK, msuMIME, msuOTHER } substate;
  MIME_Struct header, body;
  bool_t prev_line_empty;
  bool_t skip_until_boundary;
  bool_t corruption_check;
  bool_t skip_header;
  bool_t strip_header_comment;
  enum { psPLAIN, psUUENCODE } plainstate;
  struct {
    int size[MAX_BOUNDARIES];
    char identifier[MAX_BOUNDARIES][MAX_BOUNDARY_BUFSIZE];
#if defined HAVE_MBRTOWC
    wchar_t w_identifier[MAX_BOUNDARIES][MAX_BOUNDARY_BUFSIZE];
#endif
    int index;
    bool_t was_end;
  } boundary;
} MBOX_State;

typedef struct {
  enum {TEXT=1, XTAG, XTAGQUOTE, XTAGDQUOTE, TAG, TAGQUOTE, TAGDQUOTE, CMNT, DISABLED} state;
  enum {ALT=1,SRC,SRC_NETLOC,SRC_NETLOC_PREFIX,SRC_NETLOC_PATH,SRC_NETLOC_SUFFIX,UNDEF} attribute;
  enum {xpDUMB=1, xpHTML, xpSMART} parser;
  enum {SCRIPT=1,STYLE,COMMENT,NOFRAMES,NOEMBED,NOSCRIPT,NOLAYER,TITLE,VISIBLE} hide;
} XML_State;

typedef enum {xmlRESET,xmlDISABLE,xmlSMART,xmlHTML,xmlDUMB,xmlUNDEF} XML_Reset;

#ifdef __cplusplus
extern "C" 
{
#endif

  /* these are defined in catfun.c */
  char *sanitize_path(char *in);
  digitized_weight_t digitize_a_weight(weight_t w);

  void init_empirical(Empirical *emp, hash_count_t dmt, hash_bit_count_t dmhb);
  void clear_empirical(Empirical *emp);
  h_item *find_in_empirical(Empirical *emp, hash_value_t id);
  score_t empirical_entropy(Empirical *emp);


  void init_category(Category *cat);
  void free_category(Category *cat);
  c_item *find_in_category(Category *cat, hash_value_t id);
  void init_purely_random_text_category(Category *cat);
  error_code_t load_category(Category *cat);

  void score_word(char *tok, token_order_t r, regex_count_t re);

  /* string hash function in jenkins.c */
  unsigned long int hash( unsigned char *k, 
			  unsigned long int length, 
			  unsigned long int initval);

  /* file format handling in fh.c */
  void init_file_handling();
  void cleanup_file_handling();

  /* common multibyte and wide char functions in mbw.c */
  bool_t good_char(char *c);
  void init_decoding_caches();
  void free_decoding_caches();
  bool_t b64_line_filter(char *line);
  char *b64_line_filter2(char *line, char *q);
  bool_t b64_line_flush(char *line, bool_t all);
  bool_t qp_line_filter(char *line);
  char *qp_line_filter2(char *line, char *q);
  bool_t qp_line_flush(char *line, bool_t all);
  bool_t mhe_line_filter(char *line);
  int extract_header_label(char *line);
  bool_t extract_mime_boundary(char *line);
  bool_t check_mime_boundary(const char *line);
  bool_t mbox_line_filter(char *line);
  bool_t plain_text_filter(char *line);
  bool_t strings1_filter(char *line);

  void xml_character_filter(char *line);
  void process_file(FILE *input, 
		    int (*line_filter)(char *),
		    void (*character_filter)(char *), 
		    void (*word_fun)(char *, token_order_t, regex_count_t), 
		    char *(*pre_line_fun)(char *),
		    void (*post_line_fun)(char *));

  void reset_mbox_line_filter();
  void reset_xml_character_filter(XML_Reset reset);
  XML_Reset select_xml_defaults(MIME_Struct *mime);

  /* probabilities in probs.c */
  double log_poisson(int k, double lambda);


#if defined HAVE_MBRTOWC
  int w_b64_code(wchar_t c);
  int w_qp_code(wchar_t c);
  bool_t w_good_char(wchar_t *c);
  void w_init_decoding_caches();
  void w_free_decoding_caches();
  bool_t w_b64_line_filter(wchar_t *line);
  wchar_t *w_b64_line_filter2(wchar_t *line, wchar_t *q);
  bool_t w_b64_line_flush(wchar_t *line, bool_t all);
  bool_t w_qp_line_filter(wchar_t *line);
  wchar_t *w_qp_line_filter2(wchar_t *line, wchar_t *q);
  bool_t w_qp_line_flush(wchar_t *line, bool_t all);
  bool_t w_mhe_line_filter(wchar_t *line);
  int w_extract_header_label(wchar_t *line);
  bool_t w_extract_mime_boundary(wchar_t *line);
  bool_t w_check_mime_boundary(const wchar_t *line);
  bool_t w_mbox_line_filter(wchar_t *line);
  bool_t w_plain_text_filter(wchar_t *line);
  bool_t w_strings1_filter(wchar_t *line);

  int wcsncasecmp(const wchar_t *s1, const wchar_t *s2, size_t n);

  void w_xml_character_filter(wchar_t *line);
  void w_process_file(FILE *input, 
		      int (*line_filter)(wchar_t *),
		      void (*character_filter)(wchar_t *), 
		      void (*word_fun)(char *, token_order_t, regex_count_t), 
		      char *(*pre_line_fun)(char *),
		      void (*post_line_fun)(char *));

#endif

#ifdef __cplusplus
}
#endif
